<?php
declare(strict_types=1);

namespace ahrefs\AhrefsSeo;

/**
 * Detect is page noindex.
 */
class Ahrefs_Seo_Noindex extends Ahrefs_Seo_Abstract_Api {

	protected const PLUGIN_YOASTSEO = 'yoast';
	protected const PLUGIN_AIOSEO4  = 'aioseo4';
	protected const PLUGIN_AIOSEO3  = 'aioseo3';
	protected const PLUGIN_RANKMATH = 'rankmath';

	/**
	 * Analyze settings of popular SEO plugin and WordPress blog option.
	 *
	 * @since 0.7.3
	 *
	 * @param array<string,string> $page_id_to_url_list Associative array, page_tax_string => url.
	 * @return array Associave array page_tax_string => int: 0 - index, 1 - noindex, -1 - error.
	 */
	public function is_noindex( array $page_id_to_url_list ) : array {
		$results = [];
		$list    = self::get_plugins_list();

		foreach ( $page_id_to_url_list as $post_tax_string => $url ) {
			// try to load noindex from SEO plugins.
			$noindex  = null;
			$post_tax = Post_Tax::create_from_string( $post_tax_string );
			if ( isset( $list[ self::PLUGIN_YOASTSEO ] ) ) {
				$noindex = $this->is_noindex_yoast( $post_tax );
			}
			if ( is_null( $noindex ) && isset( $list[ self::PLUGIN_AIOSEO4 ] ) ) {
				$noindex = $this->is_noindex_aioseo4( $post_tax );
			}
			if ( is_null( $noindex ) && isset( $list[ self::PLUGIN_AIOSEO3 ] ) ) {
				$noindex = $this->is_noindex_aioseo3( $post_tax );
			}
			if ( is_null( $noindex ) && isset( $list[ self::PLUGIN_RANKMATH ] ) ) {
				$noindex = $this->is_noindex_rankmath( $post_tax );
			}
			if ( is_null( $noindex ) ) {
				// is blog marked as not being public?
				$noindex = '0' === (string) get_option( 'blog_public' ) ? 1 : 0;
			}

			$results[ $post_tax_string ] = $noindex;
		}
		return $results;
	}

	/**
	 * Get supported plugins list.
	 *
	 * @since 0.8.2
	 *
	 * @return array<string,string> List of [plugin => version].
	 */
	public static function get_plugins_list() : array {
		static $result = null;
		if ( is_null( $result ) ) {
			$result = [];
			if ( function_exists( 'YoastSEO' ) && defined( 'WPSEO_VERSION' ) ) {
				$result[ self::PLUGIN_YOASTSEO ] = WPSEO_VERSION;
			}
			if ( function_exists( 'aioseo' ) ) {
				$result[ self::PLUGIN_AIOSEO4 ] = property_exists( aioseo(), 'version' ) ? aioseo()->version : '4.unknown';
			} elseif ( isset( $GLOBALS['aiosp'] ) && is_object( $GLOBALS['aiosp'] ) && class_exists( '\\All_in_One_SEO_Pack' ) && ( $GLOBALS['aiosp'] instanceof \All_in_One_SEO_Pack ) ) {
				$result[ self::PLUGIN_AIOSEO3 ] = property_exists( $GLOBALS['aiosp'], 'version' ) ? $GLOBALS['aiosp']->version : '3.unknown';
			}
			if ( class_exists( '\\RankMath' ) ) {
				$result[ self::PLUGIN_RANKMATH ] = defined( 'RANK_MATH_VERSION' ) ? RANK_MATH_VERSION : 'unknown';
			}
		}

		return $result;
	}

	/**
	 * Try to get noindex value from Yoast SEO plugin, version 14+
	 *
	 * @since 0.8.2
	 *
	 * @param Post_Tax $post_tax Post or term.
	 * @return int Result, 1 - noindex, 0 - index, null - unknown.
	 */
	protected function is_noindex_yoast( Post_Tax $post_tax ) : ?int {
		$result = null;
		try {
			if ( function_exists( 'YoastSEO' ) && defined( 'WPSEO_VERSION' ) && version_compare( WPSEO_VERSION, '14.0' ) >= 0 && class_exists( '\\Yoast\\WP\\SEO\\Surfaces\\Meta_Surface' ) ) { // v14.0 and later.
				$robots_array = null;
				/** @var \Yoast\WP\SEO\Surfaces\Meta_Surface $meta */
				$meta   = YoastSEO()->meta; // @phpstan-ignore-line -- Plugin may be active or not exists.
				$values = $post_tax->is_post() ? $meta->for_post( $post_tax->get_post_id() ) : $meta->for_term( $post_tax->get_post_id() );
				if ( $values ) {
					$robots_array = $values->robots;
				}
				if ( ! is_null( $robots_array ) ) {
					$result = in_array( 'noindex', $robots_array, true ) ? 1 : 0;
					Ahrefs_Seo::breadcrumbs( sprintf( '%s(%s): %d', __METHOD__, (string) $post_tax, $result ) );
				}
			}
		} catch ( \Exception $e ) {
			Ahrefs_Seo::notify( new Ahrefs_Seo_Exception( sprintf( 'Error in %s(%s)', __METHOD__, (string) $post_tax ), 0, $e ) );
		} catch ( \Error $e ) {
			Ahrefs_Seo::notify( new Ahrefs_Seo_Exception( sprintf( 'Error in %s(%s)', __METHOD__, (string) $post_tax ), 0, $e ) );
		}

		return $result;
	}

	// phpcs:disable WordPress.WP.GlobalVariablesOverride.Prohibited,WordPress.WP.DiscouragedFunctions.wp_reset_query_wp_reset_query
	/**
	 * Try to get noindex value from All in One SEO plugin version 4.0+
	 *
	 * @since 0.8.2
	 *
	 * @param Post_Tax $post_tax Post or term.
	 * @return int Result, 1 - noindex, 0 - index, null - unknown.
	 */
	protected function is_noindex_aioseo4( Post_Tax $post_tax ) : ?int {
		$result = null;
		try {
			if ( function_exists( 'aioseo' ) && class_exists( '\\AIOSEO\\Plugin\\AIOSEO' )
			&& property_exists( aioseo(), 'db' ) ) { // @phpstan-ignore-line -- Plugin may be active or not exists.
				if ( $post_tax->is_post() ) {
					$the_post = aioseo()->db // @phpstan-ignore-line -- Plugin may be active or not exists.
					->start( 'aioseo_posts' )
					->where( 'post_id', $post_tax->get_post_id() )
					->run()
					->model( 'AIOSEO\\Plugin\\Common\\Models\\Post' );

					if ( is_object( $the_post ) && property_exists( $the_post, 'robots_noindex' ) ) {
						$result = $the_post->robots_noindex ? 1 : 0;
						Ahrefs_Seo::breadcrumbs( sprintf( '%s(%s): %d', __METHOD__, (string) $post_tax, $result ) );
					}
				} else {
					// one option for all terms of whole taxonomy.
					$args = $this->get_post_tax_args( $post_tax );
					if ( is_array( $args ) && class_exists( '\\AIOSEO\\Plugin\\Common\\Meta\\Robots' ) ) {
						global $wp_query;
						$old_wp_query = $wp_query;
						wp_reset_query();

						// define new query.
						$wp_query = new \WP_Query( $args );
						try {
							// create robots instance.
							$robots        = new \AIOSEO\Plugin\Common\Meta\Robots();
							$robots_string = false;
							// remove just added actions.
							if ( is_object( $robots ) && method_exists( $robots, 'meta' ) ) {
								if ( method_exists( $robots, 'noindexFeed' ) ) {
									remove_action( 'template_redirect', [ $robots, 'noindexFeed' ] );
								}
								if ( method_exists( $robots, 'disableWpRobotsCore' ) ) {
									remove_action( 'wp_head', [ $robots, 'disableWpRobotsCore' ] );
								}
								$robots_string = $robots->meta(); // string or false.
							}

							if ( is_string( $robots_string ) ) {
								$result = ( false !== stripos( $robots_string, 'noindex' ) ) ? 1 : 0;
								Ahrefs_Seo::breadcrumbs( sprintf( '%s(%s): %d', __METHOD__, (string) $post_tax, $result ) );
							}
						} finally {
							$GLOBALS['wp_query'] = $old_wp_query; // phpcs:ignore WordPress.WP.GlobalVariablesOverride.Prohibited -- restore the original query.
						}
					}
				}
			}
		} catch ( \Exception $e ) {
			Ahrefs_Seo::notify( new Ahrefs_Seo_Exception( sprintf( 'Error in %s(%s)', __METHOD__, (string) $post_tax ), 0, $e ) );
		} catch ( \Error $e ) {
			Ahrefs_Seo::notify( new Ahrefs_Seo_Exception( sprintf( 'Error in %s(%s)', __METHOD__, (string) $post_tax ), 0, $e ) );
		}
		return $result;
	}

	/**
	 * Try to get noindex value from All in One SEO plugin, version 3.3.5 - 3.7.1
	 *
	 * @since 0.8.2
	 *
	 * @param Post_Tax $post_tax Post or term.
	 * @return int Result, 1 - noindex, 0 - index, null - unknown.
	 */
	protected function is_noindex_aioseo3( Post_Tax $post_tax ) : ?int {
		global $aiosp;
		$result = null;
		$args   = $this->get_post_tax_args( $post_tax );
		try {
			if ( is_array( $args ) && ! empty( $aiosp ) && class_exists( '\\All_in_One_SEO_Pack' )
			&& ( $aiosp instanceof \All_in_One_SEO_Pack ) && class_exists( '\\AIOSEOP_Robots_Meta' ) && method_exists( '\\AIOSEOP_Robots_Meta', 'get_robots_meta_tag' ) ) {
				global $wp_query;
				$old_wp_query = $wp_query;
				wp_reset_query();

				// define new query.
				$wp_query = new \WP_Query( $args );
				try {
					$aioseop_robots_meta = new \AIOSEOP_Robots_Meta();
					$robots_meta_string  = $aioseop_robots_meta->get_robots_meta_tag();
					if ( is_string( $robots_meta_string ) ) {
						$result = false !== stripos( $robots_meta_string, 'noindex' ) ? 1 : 0;
						Ahrefs_Seo::breadcrumbs( sprintf( '%s(%s): %d', __METHOD__, (string) $post_tax, $result ) );
					}
				} finally {
					$GLOBALS['wp_query'] = $old_wp_query; // phpcs:ignore WordPress.WP.GlobalVariablesOverride.Prohibited -- restore the original query.
				}
			}
		} catch ( \Exception $e ) {
			Ahrefs_Seo::notify( new Ahrefs_Seo_Exception( sprintf( 'Error in %s(%s)', __METHOD__, (string) $post_tax ), 0, $e ) );
		} catch ( \Error $e ) {
			Ahrefs_Seo::notify( new Ahrefs_Seo_Exception( sprintf( 'Error in %s(%s)', __METHOD__, (string) $post_tax ), 0, $e ) );
		}
		return $result;
	}
	// phpcs:enable WordPress.WP.GlobalVariablesOverride.Prohibited,WordPress.WP.DiscouragedFunctions.wp_reset_query_wp_reset_query

	/**
	 * Try to get noindex value from Rank Math SEO plugin, version 1.0.0+
	 *
	 * @since 0.8.2
	 *
	 * @param Post_Tax $post_tax Post or term.
	 * @return int Result, 1 - noindex, 0 - index, null - unknown.
	 */
	protected function is_noindex_rankmath( Post_Tax $post_tax ) : ?int {
		$result = null;
		try {
			if ( class_exists( '\\RankMath\\Helper' )
			&& method_exists( '\\RankMath\\Helper', 'is_post_indexable' ) && method_exists( '\\RankMath\\Helper', 'is_term_indexable' ) ) { // @phpstan-ignore-line -- Plugin may be active or not exists.
				if ( $post_tax->is_post() ) {
					$result = \RankMath\Helper::is_post_indexable( $post_tax->get_post_id() ) ? 0 : 1;
				} else {
					$term = get_term_by( 'id', $post_tax->get_post_id(), $post_tax->get_taxonomy() );
					if ( $term instanceof \WP_Term ) {
						$result = \RankMath\Helper::is_term_indexable( $term ) ? 0 : 1;
					}
				}
			}
		} catch ( \Exception $e ) {
			Ahrefs_Seo::notify( new Ahrefs_Seo_Exception( sprintf( 'Error in %s(%s)', __METHOD__, (string) $post_tax ), 0, $e ) );
		} catch ( \Error $e ) {
			Ahrefs_Seo::notify( new Ahrefs_Seo_Exception( sprintf( 'Error in %s(%s)', __METHOD__, (string) $post_tax ), 0, $e ) );
		}
		return $result;
	}

	/**
	 * Get wp_query options for item
	 *
	 * @since 0.8.2
	 *
	 * @param Post_Tax $post_tax Post or term.
	 * @return array<string, string|int|array>|null Arguments for wp_query.
	 */
	private function get_post_tax_args( Post_Tax $post_tax ) : ?array {
		if ( $post_tax->is_post() ) {
			return [
				'p'              => $post_tax->get_post_id(),
				'post_status'    => 'publish',
				'post_type'      => ( new Ahrefs_Seo_Content() )->get_custom_post_types_enabled( true ),
				'posts_per_page' => 1,
			];
		}
		$term = get_term_by( 'id', $post_tax->get_post_id(), $post_tax->get_taxonomy() );
		if ( is_object( $term ) && ( $term instanceof \WP_Term ) ) {
			if ( 'category' === $term->taxonomy ) {
				return [ 'category_name' => $term->slug ];
			} else {
				return [ $term->taxonomy => $term->term_id ];
			}
		}
		return null;
	}

}
