<?php
declare(strict_types=1);
namespace ahrefs\AhrefsSeo;

$content = new Ahrefs_Seo_Content_Settings();

if ( ! function_exists( 'render_scope_block' ) ) {
	/**
	 * Display scope block for post type
	 *
	 * @since 0.8.0
	 *
	 * @param string   $block_title Title of whole options block.
	 * @param string   $main_item_title Main item title.
	 * @param string   $main_checkbox_name Value of main checkbox HTML name attribute.
	 * @param bool     $main_checkbox_is_checked Is main checkbox checked.
	 * @param int      $subitems_count Total subitems count.
	 * @param bool     $hide_by_count True: show only first 10 subitems, hide the rest; False: set visible items using block height.
	 * @param callable $render_subitems Function for subitems render.
	 * @return void
	 */
	function render_scope_block( string $block_title, string $main_item_title, string $main_checkbox_name, bool $main_checkbox_is_checked, int $subitems_count, bool $hide_by_count, callable $render_subitems ) : void {
		$show_more = $subitems_count > 10;
		?>
		<div class="scope-block">
			<div class="block-title"><?php echo esc_html( $block_title ); ?></div>
			<div class="one-column">
				<ul class="checkbox-group<?php echo $show_more ? ' group-expandable' : ''; ?><?php echo $hide_by_count ? ' hide-by-count' : ' hide-by-height'; ?>">
					<li>
						<label>
							<input type="checkbox" value="1" name="<?php echo esc_attr( $main_checkbox_name ); ?>" class="checkbox-main" <?php checked( $main_checkbox_is_checked ); ?>>
							<?php
							echo esc_html( $main_item_title );
							?>
						</label>
						<ul class="subitems">
							<?php $render_subitems(); ?>
						</ul>
					</li>
					<li class="show-more-li"><a class="show-more-a" href="#"><?php esc_html_e( 'Show more', 'ahrefs-seo' ); ?></a></li>
					<li class="show-less-li"><a class="show-more-a" href="#"><?php esc_html_e( 'Show less', 'ahrefs-seo' ); ?></a></li>
				</ul>
			</div>
		</div>
		<hr class="hr-shadow">
		<?php
	}
}

$pages_disabled = $content->is_disabled_for_pages();
$posts_disabled = $content->is_disabled_for_posts();

$pages_list = $content->get_pages_list();
// By default, we should tick all the boxes for 'Pages' & 'Posts' on the set up wizard page.
$posts_checked = $content->get_posts_categories_checked(); // load from options.
if ( is_null( $posts_checked ) ) {
	$posts_checked = Helper_Content::get()->get_all_term_ids( 'category' );
}
$page_ids = $content->get_pages_checked();
if ( is_null( $page_ids ) ) {
	$page_ids       = []; // check nothing by default.
	$pages_disabled = true;
}
?>

<div class="block-text">
	<?php esc_html_e( 'Uncheck the categories below that don’t need organic search performance improvement, like "Contacts" or "Privacy Policy". This will give you a more accurate analysis and better recommendations.', 'ahrefs-seo' ); ?>
</div>

<?php
$args             = [
	'fields'     => 'ids',
	'hide_empty' => false,
];
$category_11_list = Helper_Content::get()->get_categories( 'category', $args, '', 11 ); // load up to 11 categories - do not display 'show more' if there are less that 10 categories.

// 1. Posts.
render_scope_block(
	__( 'Posts', 'ahrefs-seo' ),
	__( 'All Posts', 'ahrefs-seo' ),
	'posts_enabled',
	! $posts_disabled,
	count( $category_11_list ),
	false,
	function() use ( $posts_checked ) {
		Helper_Content::get()->terms_checklists( 'category', $posts_checked );
	}
);

// 2. Pages.
render_scope_block(
	__( 'Pages', 'ahrefs-seo' ),
	__( 'All Pages', 'ahrefs-seo' ),
	'pages_enabled',
	! $pages_disabled,
	count( $pages_list ),
	true,
	function() use ( $pages_list, $page_ids ) {
		$i = 0;
		foreach ( $pages_list as $id => $title ) {
			?>
			<li>
				<label class="selectit">
					<input type="checkbox" name="pages[]" value="<?php echo esc_attr( "$id" ); ?>" id="<?php echo esc_attr( 'page_' . $id ); ?>" <?php checked( in_array( "$id", $page_ids, true ) ); ?><label for="<?php echo esc_attr( 'page_' . $id ); ?>"><?php echo esc_html( $title ); ?></label>
				</label><span class="height26px"></span>
			</li>
			<?php
		}
	}
);

?>
<a id="cpt"></a>
<?php

// 3. Custom post types.
$all_cpt_disabled          = $content->is_disabled_for_all_cpt();
$custom_post_types         = $content->get_custom_post_types();
$custom_post_types_enabled = $content->get_custom_post_types_enabled();
if ( count( $custom_post_types ) ) {
	render_scope_block(
		__( 'Custom Post Types', 'ahrefs-seo' ),
		__( 'All Custom Post Types', 'ahrefs-seo' ),
		'all_cpt_enabled',
		! $all_cpt_disabled,
		count( $custom_post_types ),
		true,
		function() use ( $custom_post_types, $custom_post_types_enabled ) {
			foreach ( $custom_post_types as $slug => $title ) {
				?>
				<li>
					<label>
						<input type="checkbox" value="<?php echo esc_attr( "$slug" ); ?>" name="cpt[]" <?php checked( in_array( $slug, $custom_post_types_enabled, true ) ); ?>>
						<?php
						echo esc_html( $title );
						?>
					</label>
				</li>
				<?php
			}
		}
	);
}
// 4. Product.
$products_exists = $content->products_exists();
if ( $products_exists ) {
	$products_disabled = $content->is_disabled_for_products();
	$products_checked  = $content->get_products_categories_checked(); // load from options.
	$args              = [
		'fields'     => 'ids',
		'hide_empty' => false,
	];

	$products_11_list = Helper_Content::get()->get_categories( 'product_cat', $args, '', 11 ); // load up to 11 categories - do not display 'show more' if there are less that 10 items.

	render_scope_block(
		__( 'Products', 'ahrefs-seo' ),
		__( 'All Product Categories', 'ahrefs-seo' ),
		'products_enabled',
		! $products_disabled,
		count( $products_11_list ),
		false,
		function() use ( $products_checked ) {
			if ( taxonomy_exists( 'product_cat' ) ) {
				$args = [
					'descendants_and_self' => 0,
					'checked_ontop'        => false,
				];
				Helper_Content::get()->terms_checklists( 'product_cat', $products_checked ?? [] );
			}
		}
	);
}
